/*
	File:		FireStarter.c

	Contains:	Code to blast other machines VRAM using FireWire.

	Written by:	Quinn "The Eskimo!", Apple Worldwide Developer Technical Support

	Copyright:	Copyright  2000-2002 by Apple Computer, Inc., All Rights Reserved.

	Disclaimer:	IMPORTANT:  This Apple software is supplied to you by Apple Computer, Inc.
				("Apple") in consideration of your agreement to the following terms, and your
				use, installation, modification or redistribution of this Apple software
				constitutes acceptance of these terms.  If you do not agree with these terms,
				please do not use, install, modify or redistribute this Apple software.

				In consideration of your agreement to abide by the following terms, and subject
				to these terms, Apple grants you a personal, non-exclusive license, under Apples
				copyrights in this original Apple software (the "Apple Software"), to use,
				reproduce, modify and redistribute the Apple Software, with or without
				modifications, in source and/or binary forms; provided that if you redistribute
				the Apple Software in its entirety and without modifications, you must retain
				this notice and the following text and disclaimers in all such redistributions of
				the Apple Software.  Neither the name, trademarks, service marks or logos of
				Apple Computer, Inc. may be used to endorse or promote products derived from the
				Apple Software without specific prior written permission from Apple.  Except as
				expressly stated in this notice, no other rights or licenses, express or implied,
				are granted by Apple herein, including but not limited to any patent rights that
				may be infringed by your derivative works or by other works in which the Apple
				Software may be incorporated.

				The Apple Software is provided by Apple on an "AS IS" basis.  APPLE MAKES NO
				WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION THE IMPLIED
				WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS FOR A PARTICULAR
				PURPOSE, REGARDING THE APPLE SOFTWARE OR ITS USE AND OPERATION ALONE OR IN
				COMBINATION WITH YOUR PRODUCTS.

				IN NO EVENT SHALL APPLE BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL OR
				CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE
				GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
				ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION, MODIFICATION AND/OR DISTRIBUTION
				OF THE APPLE SOFTWARE, HOWEVER CAUSED AND WHETHER UNDER THEORY OF CONTRACT, TORT
				(INCLUDING NEGLIGENCE), STRICT LIABILITY OR OTHERWISE, EVEN IF APPLE HAS BEEN
				ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

	Change History (most recent first):

*/

#define __CF_USE_FRAMEWORK_INCLUDES__ 1

#include <Carbon/Carbon.h>
#include <QuickTime/QuickTime.h>
#include <IOKit/IOKitLib.h>
#include <IOKit/firewire/IOFireWireLib.h>
#include <IOKit/IOMessage.h>

#include <stdio.h>
#include <assert.h>

struct FWDeviceInfo {
	long long guid;
	UInt32 baseAddr;
	int    columns;
	int	   rows;
	int	   rowBytes;
};
typedef struct FWDeviceInfo FWDeviceInfo;

static const FWDeviceInfo gDevices[] = {
	{	// Strawberry iMac for demo
		13622949040548594,
		0x94008000,
		1024,
		768,
		1024 * 4
	},
	{	// Flower Power iMac for demo
		13622949046945768,
		0x96008000,
		1024,
		768,
		1024 * 4
	},
	{	// QuickSilver G4 in Lab
		1007152618436672,
		0x98004000,
		1280,
		1024,
		1280 * 4
	},
	{	// Flower Power iMac in Lab
		13622949046924916,
		0x96000000 + 8 * 4096,
		800,
		600,
		832 * 4
	},
	{	// Meeroh's TiBook
		13622949048632576,
		0xb4000000, // + 8 * 4096,
		1152,
		768,
		1152 * 4
	},
	{	// iMac flat panel in Lab
		1007152625873848,
		0x98004000,
		1024,
		768,
		1024 * 4
	}
};

// Chris Russ, 800 MHz TiBook, running 9, read 0 works
// Meeroh, 500 MHz TiBook, running X, read 0 fails
// iMac Flower Power, running X, everything works
// Peter Sichel, iBook 600 MHz, read 0 works

static io_service_t  gCurrentService = MACH_PORT_NULL;
static io_object_t   gCurrentServiceNotification = MACH_PORT_NULL;
static io_iterator_t gIterator = MACH_PORT_NULL;
static IOCFPlugInInterface** 	gPlugIn = NULL;
static IOFireWireLibDeviceRef	gFWIntf = NULL;
static int gDeviceIndex;

static ImageSequence			gImageSequence;
static QTAtomContainer			gEffectDescription;
static TimeBase					gTimeBase;
static ImageDescriptionHandle	gSampleDescription;
static WindowRef				gMainWindow;

static int gNumberOfSteps = 30;

// kFireCodecType

static ImageDescriptionHandle QTEffects_MakeSampleDescription(OSType theEffectType, short theWidth, short theHeight)
{
	ImageDescriptionHandle		mySampleDesc = NULL;

	OSErr						myErr = noErr;
	
	// create a new sample description
	myErr = MakeImageDescriptionForEffect(theEffectType, &mySampleDesc);
	if (myErr != noErr)
		return(NULL);
	
	(**mySampleDesc).vendor = kAppleManufacturer;
	(**mySampleDesc).temporalQuality = codecNormalQuality;
	(**mySampleDesc).spatialQuality = codecNormalQuality;
	(**mySampleDesc).width = theWidth;
	(**mySampleDesc).height = theHeight;
	
	return(mySampleDesc);
}

// kFireCodecType

static QTAtomContainer QTEffects_CreateEffectDescription (OSType theEffectName)
{
	QTAtomContainer		myEffectDesc = NULL;
	OSType				myType;
	OSErr				myErr = noErr;

	// create a new, empty effect description
	myErr = QTNewAtomContainer(&myEffectDesc);
	if (myErr != noErr)
		goto bail;

	// create the effect ID atom: the atom type is kParameterWhatName, and the atom ID is kParameterWhatID
	myType = EndianU32_NtoB(theEffectName);
	myErr = QTInsertChild(myEffectDesc, kParentAtomIsContainer, kParameterWhatName, kParameterWhatID, 0, sizeof(myType), &myType, NULL);
	if (myErr != noErr)
		goto bail;
		
bail:
	return(myEffectDesc);
}

static OSErr QTEffects_SetUpEffectSequence (void)
{
	OSErr						myErr = noErr;
 	
	// make an effects sequence
	HLock((Handle)gEffectDescription);

	// prepare the decompression sequence for playback
	myErr = DecompressSequenceBeginS(
							&gImageSequence,
							gSampleDescription,
							*gEffectDescription,
							GetHandleSize(gEffectDescription),
							(CGrafPtr)GetWindowPort(gMainWindow),
							NULL,
							NULL,
							NULL,
							ditherCopy,
							NULL,
							0,
							codecNormalQuality,
							NULL);

	HUnlock((Handle)gEffectDescription);
	if (myErr != noErr)
		goto bail;

	// create a new time base and associate it with the decompression sequence
	gTimeBase = NewTimeBase();
	myErr = GetMoviesError();
	if (myErr != noErr)
		goto bail;

	SetTimeBaseRate(gTimeBase, 0);
	myErr = CDSequenceSetTimeBase(gImageSequence, gTimeBase);

bail:
	return(myErr);
}

static OSErr QTEffects_RunEffect(TimeValue theTime)
{
	OSErr						myErr = noErr;
	ICMFrameTimeRecord			myFrameTime;

	// set the timebase time to the step of the sequence to be rendered
	SetTimeBaseValue(gTimeBase, theTime, gNumberOfSteps);

	myFrameTime.value.hi				= 0;
	myFrameTime.value.lo				= theTime;
	myFrameTime.scale					= gNumberOfSteps;
	myFrameTime.base					= 0;
	myFrameTime.duration				= gNumberOfSteps;
	myFrameTime.rate					= 0;
	myFrameTime.recordSize				= sizeof(myFrameTime);
	myFrameTime.frameNumber				= 1;
	myFrameTime.flags					= icmFrameTimeHasVirtualStartTimeAndDuration;
	myFrameTime.virtualStartTime.lo		= 0;
	myFrameTime.virtualStartTime.hi		= 0;
	myFrameTime.virtualDuration			= gNumberOfSteps;
	
	HLock((Handle)gEffectDescription);

	myErr = DecompressSequenceFrameWhen(
										gImageSequence,
										*((Handle)gEffectDescription),
										GetHandleSize((Handle)gEffectDescription),
										0,
										0,
										NULL,
										&myFrameTime);
										
	HUnlock((Handle)gEffectDescription);
	
	if (myErr != noErr)
		goto bail;
	
bail:
	return(myErr);
}

extern pascal void CFQRelease(CFTypeRef cf);
extern pascal void CFQRelease(CFTypeRef cf)
	// See comment in header.
{
	if (cf != nil) {
		CFRelease(cf);
	}
}

extern pascal OSStatus CFQDictionarySetNumber(CFMutableDictionaryRef dict, const void *key, long long value);
extern pascal OSStatus CFQDictionarySetNumber(CFMutableDictionaryRef dict, const void *key, long long value)
	// See comment in header.
{
	OSStatus    err;
	CFNumberRef valueNum;

	// Create a CFNumber and add it to the dictionary.
		
	err = noErr;
	valueNum = CFNumberCreate(nil, kCFNumberLongLongType, &value);
	if (valueNum == nil) {
		err = coreFoundationUnknownErr;
	}
	if (err == noErr) {
		CFDictionarySetValue(dict, key, valueNum);
	}
	CFQRelease(valueNum);
	
	return err;
}

static mach_port_t gMasterPort;
static IONotificationPortRef gNotificationPortRef;

enum {
    kCSRRegisterSpaceBaseAddressHi	= 0x0000FFFF,
    kCSRRegisterSpaceBaseAddressLo	= 0xF0000000,
    kConfigROMBaseAddress			= kCSRRegisterSpaceBaseAddressLo + 0x0400
};

enum {
	kChunkSize = 1024
};

static char gZeroes[kChunkSize];

static void EraseScreen(void)
{
	OSStatus err;
	UInt32 dstBase;
	int dstRowBytes;
	int dstColumns;
	int dstRows;
	int rowIndex;
	int colIndex;
	FWAddress 	addr;
	UInt32 size;

	dstBase     = gDevices[gDeviceIndex].baseAddr;
	dstRowBytes = gDevices[gDeviceIndex].rowBytes;
	dstColumns  = gDevices[gDeviceIndex].columns;
	dstRows     = gDevices[gDeviceIndex].rows;

	for (rowIndex = 0; rowIndex < dstRows; rowIndex++) {
		int numCols;

		numCols = ((dstColumns * 4 + kChunkSize - 1) / kChunkSize);

		// fprintf(stderr, "  rowIndex = %d, numCols = %d\n", rowIndex, numCols);
		
		for (colIndex = 0; colIndex < numCols; colIndex++) {
			addr.addressHi = 0;
			addr.addressLo = dstBase + (dstRowBytes * rowIndex) + (colIndex * kChunkSize);
			size = kChunkSize;
			if (colIndex == (numCols - 1)) {
				size = (dstColumns * 4) - (colIndex * kChunkSize);
			}
			err = (*gFWIntf)->Write(gFWIntf, gCurrentService, 
									&addr, 
									gZeroes,
									&size, 
									false, 0);

			if (err == 0) {
				// fprintf(stderr, "  size = %ld\n", size);
			} else {
				break;
			}
		}
		if (err != 0) {
			break;
		}
	}
	if (err != noErr) {
		fprintf(stderr, "EraseScreen: err = %08lx, rowIndex = %d\n", err, rowIndex);
	}
}

static Boolean gFirstTime = true;

static void BlastScreen(void)
{
	OSStatus err;
	OSStatus junk;
	int rowIndex;
	int colIndex;
	FWAddress 	addr;
	UInt32 size;
	PixMapHandle pm;
	Ptr srcBaseAddr;
	long srcRowBytes;
	UInt32 dstBase;
	int dstRowBytes;
	Rect srcBounds;
	int dstColumns;
	int dstRows;
	char tmpBuffer[kChunkSize];
	static long extra = 0;

	if (gFirstTime) {
		EraseScreen();
		gFirstTime = false;
	}

	#if 1
		junk = LockPortBits(GetWindowPort(gMainWindow));
		assert(junk == noErr);
		pm = GetPortPixMap(GetWindowPort(gMainWindow));
	#else
		pm = (**GetMainDevice()).gdPMap;
	#endif
	(void) GetPixBounds(pm, &srcBounds);
	srcRowBytes = GetPixRowBytes(pm);
	srcBaseAddr = GetPixBaseAddr(pm);
	// fprintf(stderr, "BlastScreen\n");
	// fprintf(stderr, "  srcBaseAddr = %08x\n", srcBaseAddr);

	dstBase     = gDevices[gDeviceIndex].baseAddr;
	dstRowBytes = gDevices[gDeviceIndex].rowBytes;
	dstColumns  = gDevices[gDeviceIndex].columns;
	dstRows     = gDevices[gDeviceIndex].rows;

	srcBaseAddr += ( ((long) -srcBounds.top) * srcRowBytes);
	srcBounds.top = 0;

	if (dstColumns > (srcBounds.right - srcBounds.left)) {
		dstBase += ((dstColumns - (srcBounds.right - srcBounds.left)) / 2) * 4;

		dstColumns = (srcBounds.right - srcBounds.left);
	}	
	if (dstRows > (srcBounds.bottom - srcBounds.top)) {
		dstBase += ((dstRows - (srcBounds.bottom - srcBounds.top)) / 2) * dstRowBytes;
		
		dstRows = (srcBounds.bottom - srcBounds.top);
	}

	
	// fprintf(stderr, "  srcBounds.top = %d\n", srcBounds.top);
	// fprintf(stderr, "  srcRowBytes   = %d\n", srcRowBytes);
	
	// fprintf(stderr, "  srcBaseAddr = %08x\n", srcBaseAddr);
	// fprintf(stderr, "  srcRowBytes = %d\n",   srcRowBytes);
	
	for (rowIndex = 0; rowIndex < dstRows; rowIndex++) {
		int numCols;

		numCols = ((dstColumns * 4 + kChunkSize - 1) / kChunkSize);

		// fprintf(stderr, "  rowIndex = %d, numCols = %d\n", rowIndex, numCols);
		
		for (colIndex = 0; colIndex < numCols; colIndex++) {
			Ptr srcAddr;
			
			addr.addressHi = 0;
			addr.addressLo = dstBase + (dstRowBytes * rowIndex) + (colIndex * kChunkSize) + extra;
			// extra += 1024 * 1024;
			size = kChunkSize;
			if (colIndex == (numCols - 1)) {
				size = (dstColumns * 4) - (colIndex * kChunkSize);
			}
			// fprintf(stderr, "  size = %ld\n", size);

			srcAddr = srcBaseAddr + (srcRowBytes * rowIndex) + (colIndex * kChunkSize), 
			// fprintf(stderr, "  srcAddr = %08x\n", srcAddr);
			BlockMoveData(srcAddr, tmpBuffer, size);
			
			// fprintf(stderr, "  addr.addressLo = %08lx\n", addr.addressLo);
			err = (*gFWIntf)->Write(gFWIntf, gCurrentService, 
									&addr, 
									tmpBuffer,
									&size, 
									false, 0);

			if (err == 0) {
				// fprintf(stderr, "  size = %ld\n", size);
			} else {
				break;
			}
		}
		if (err != 0) {
			break;
		}
	}

	if (err != noErr) {
		fprintf(stderr, "  err = %08lx, rowIndex = %d\n", err, rowIndex);
	} else {
		// fprintf(stderr, "  success\n");
	}
	
	junk = UnlockPortBits(GetWindowPort(gMainWindow));
	assert(junk == noErr);
}

static pascal OSErr QuitAppleEventHandler(const AppleEvent *appleEvt, AppleEvent* reply, long refCon)
{
	#pragma unused (appleEvt)
	#pragma unused (reply)
	#pragma unused (refCon)

	QuitApplicationEventLoop();
	
	return noErr;
}

static pascal void TimerCallback(EventLoopTimerRef inTimer, void *inUserData)
{
	OSStatus err;
	#pragma unused(inTimer)
	#pragma unused(inUserData)
	
	err = QTEffects_RunEffect(1);
	assert(err == noErr);
	
	if (gFWIntf != nil) {
		BlastScreen();
	}
}

static char gBuffer[1024];

static void ScanFireWireDevices(void)
{
	OSStatus 				err;
	OSStatus 				junk;
	CFMutableDictionaryRef 	matchDict;
	io_iterator_t			iterator;
	
	iterator = MACH_PORT_NULL;
	
	err = noErr;
	matchDict = IOServiceMatching("IOFireWireDevice");
	if (matchDict == NULL) {
		err = -1;
	}
	
	if (err == noErr) {
		err = IOServiceGetMatchingServices(gMasterPort, matchDict, &iterator);
		matchDict = nil;
	}

	if (err == noErr) {
		int deviceNumber;
		Boolean done;
		
		deviceNumber = 0;
		do {
			io_service_t 	thisService;
			
			thisService = IOIteratorNext(iterator);
			done = (thisService == MACH_PORT_NULL);
			if ( ! done ) {
				IOCFPlugInInterface** 	plugIn;
				IOFireWireLibDeviceRef	fwIntf;
				UInt32 val;
				FWAddress 	addr;
				SInt32 junkScore;

				fprintf(stderr, "deviceNumber = %d\n", deviceNumber);
				
				plugIn = nil;
				fwIntf = nil;
				
				err = IOCreatePlugInInterfaceForService(thisService, kIOFireWireLibTypeID, kIOCFPlugInInterfaceID, &plugIn, &junkScore);
				if (err == 0) {
					err = (*plugIn)->QueryInterface(plugIn, CFUUIDGetUUIDBytes(kIOFireWireDeviceInterfaceID), (void **) &fwIntf);
				}
				if (err == 0) {
					assert(	(*fwIntf)->InterfaceIsInited(fwIntf) );
					
					err = (*fwIntf)->Open(fwIntf);
				}
				if (err == 0) {
					addr.nodeID = 0;
					addr.addressHi = kCSRRegisterSpaceBaseAddressHi;
					addr.addressLo = kConfigROMBaseAddress;

					err = (*fwIntf)->ReadQuadlet(fwIntf, thisService, &addr, &val, false, 0);
					if (err == 0) {
						fprintf(stderr, "  val[CSR] = %lx\n", val);
					}
				}
				if (err == 0) {
					addr.addressHi = 0;
					addr.addressLo = 0;
					
					err = (*fwIntf)->ReadQuadlet(fwIntf, thisService, &addr, &val, false, 0);
					if (err == 0) {
						fprintf(stderr, "  val[0] = %lx\n", val);
					}
				}
				if (err == 0 && 1) {
					UInt32 size;
					
					addr.addressHi = 0;
					addr.addressLo = 0x94000000;
					
					size = sizeof(gBuffer);
					err = (*fwIntf)->Write(fwIntf, thisService, &addr, gBuffer, &size, false, 0);
				}

				if (fwIntf != nil) {
					(*fwIntf)->Close(fwIntf);
					// how to free fwIntf
					gFWIntf = nil;
				}
				if (plugIn != nil) {
					junk = IODestroyPlugInInterface(plugIn);
					assert(junk == noErr);
					plugIn = nil;
				}
			}
			
			if (thisService != MACH_PORT_NULL) {
				junk = IOObjectRelease(thisService);
				assert(junk == 0);
			}
			if (err != noErr) {
				fprintf(stderr, "  err = %lx\n", err);
				err = noErr;
			}
			deviceNumber += 1;
		} while ( ! done );
	}

	CFQRelease(matchDict);
	if (iterator != MACH_PORT_NULL) {
		junk = IOObjectRelease(iterator);
		assert(junk == noErr);
	}
		
	fprintf(stderr, "  err = %lx\n", err);
}

static pascal OSStatus CommandHandler(EventHandlerCallRef inHandlerCallRef, EventRef inEvent, void *inUserData)
{
	OSStatus 	err;
	HICommand 	thisCommand;
	#pragma unused(inHandlerCallRef)
	#pragma unused(inUserData)
	
	assert( GetEventClass(inEvent) == kEventClassCommand );
	assert( GetEventKind(inEvent)  == kEventCommandProcess );

	err = GetEventParameter(inEvent, kEventParamDirectObject, 
							typeHICommand, nil, 
							sizeof(thisCommand), nil, &thisCommand);
	if (err == noErr) {
		switch (thisCommand.commandID) {
			case 'SCAN':
				fprintf(stderr, "SCAN\n");
				ScanFireWireDevices();
				break;
			default:
				err = eventNotHandledErr;
				break;
		}
	}

	return err;
}

static OSStatus Initialize(void)
{
    OSStatus 		err;
    Rect     		bounds;
    IBNibRef 		nibRef;
       
    nibRef = nil;
    
    InitCursor();

	err = EnterMovies();
    if (err == noErr) {
	    err = AEInstallEventHandler(kCoreEventClass, kAEQuitApplication, NewAEEventHandlerUPP(QuitAppleEventHandler), 0, false);
	}
    if (err == noErr) {
		bounds.top = 100;
		bounds.left = 100;
		bounds.bottom = 200;
		bounds.right = 700;

		err = CreateNewWindow(kDocumentWindowClass, 
							  kWindowCloseBoxAttribute | kWindowStandardHandlerAttribute,
                       		  &bounds, 
                       		  &gMainWindow);
	}	
    if (err == noErr) {
    	SetWTitle(gMainWindow, "\pFireStarter");
    	ShowWindow(gMainWindow);
		err = InstallStandardEventHandler(GetApplicationEventTarget());
	}
	if (err == noErr) {
		gEffectDescription = QTEffects_CreateEffectDescription(kFireCodecType);
		if (gEffectDescription == NULL) {
			err = -1;
		}
	}
	if (err == noErr) {
		gSampleDescription = QTEffects_MakeSampleDescription(kFireCodecType, bounds.right - bounds.left, bounds.bottom - bounds.top);
		if (gSampleDescription == NULL) {
			err = -1;
		}
	}		
	if (err == noErr) {
		err = QTEffects_SetUpEffectSequence();
	}
	if (err == noErr) {
		EventLoopTimerRef junkTimerRef;

		err = InstallEventLoopTimer(GetCurrentEventLoop(), 0.1, 0.1, NewEventLoopTimerUPP(TimerCallback), nil, &junkTimerRef);
	}
	if (err == noErr) {
		err = CreateNibReference(CFSTR("Main"), &nibRef);
	}
	if (err == noErr) {
		err = SetMenuBarFromNib(nibRef, CFSTR("MenuBar"));
	}
	if (err == noErr) {
		static const EventTypeSpec kEventTypes = {kEventClassCommand, kEventCommandProcess};
		
		err = InstallEventHandler(GetApplicationEventTarget(), NewEventHandlerUPP(CommandHandler), 1, &kEventTypes, nil, nil);
	}

	if (nibRef != nil) {
		DisposeNibReference(nibRef);
	}
	
    return err;
}

static void MyDeviceRemovedCallback(void *refCon, 
									io_service_t service, 
									natural_t messageType,
									void *messageArgument);

static OSStatus FoundDevice(io_service_t thisService, int deviceIndex)
{
	OSStatus 				err;
	OSStatus 				junk;
	SInt32					junkScore;
	IOCFPlugInInterface** 	plugIn;
	IOFireWireLibDeviceRef	fwIntf;
	io_object_t				thisServiceNotification;
	
	fprintf(stderr, "FoundDevice %d\n", deviceIndex);
	fprintf(stderr, "  thisService refcount = %d\n", IOObjectGetRetainCount(thisService));
	
	plugIn = nil;
	fwIntf = nil;
	
	err = IOCreatePlugInInterfaceForService(thisService, kIOFireWireLibTypeID, kIOCFPlugInInterfaceID, &plugIn, &junkScore);
	if (err == 0) {
		err = (*plugIn)->QueryInterface(plugIn, CFUUIDGetUUIDBytes(kIOFireWireDeviceInterfaceID), (void **) &fwIntf);
	}
	if (err == 0) {
		assert(	(*fwIntf)->InterfaceIsInited(fwIntf) );
		
		err = (*fwIntf)->Open(fwIntf);
	}
	if (err == 0) {
		FWAddress 	addr;
		UInt32		val;
		
		addr.nodeID = 0;
		addr.addressHi = kCSRRegisterSpaceBaseAddressHi;
		addr.addressLo = kConfigROMBaseAddress;

		err = (*fwIntf)->ReadQuadlet(fwIntf, thisService, &addr, &val, false, 0);
		if (err == 0) {
			fprintf(stderr, "  val = %lx\n", val);
		}
	}
	if (err == noErr) {	
		err = IOServiceAddInterestNotification(gNotificationPortRef, 
											   thisService, 
											   kIOGeneralInterest,
											   MyDeviceRemovedCallback,
											   nil,
											   &thisServiceNotification);
	}
	if (err == noErr) {
		gPlugIn = plugIn;
		gFWIntf = fwIntf;
		gCurrentServiceNotification = thisServiceNotification;
		gDeviceIndex = deviceIndex;
	} else {
		if (fwIntf != nil) {
			(*fwIntf)->Close(fwIntf);
			// how to free fwIntf
		}
		if (plugIn != nil) {
			junk = IODestroyPlugInInterface(plugIn);
			assert(junk == noErr);
		}
	}
	return err;
}

static void MyDeviceRemovedCallback(void *refCon, 
									io_service_t service, 
									natural_t messageType,
									void *messageArgument)
{
	OSStatus junk;
	#pragma unused(refCon)
	#pragma unused(messageArgument)
	
	assert(service == gCurrentService);
	if (service == gCurrentService) {
		if (messageType == kIOMessageServiceIsTerminated) {
			fprintf(stderr, "MyDeviceRemovedCallback\n");
			fprintf(stderr, "  gCurrentService refcount = %d\n", IOObjectGetRetainCount(gCurrentService));

			if (gFWIntf != nil) {
				(*gFWIntf)->Close(gFWIntf);
				// how to free gFWIntf
				gFWIntf = nil;
			}
			if (gPlugIn != nil) {
				junk = IODestroyPlugInInterface(gPlugIn);
				assert(junk == noErr);
				gPlugIn = nil;
			}

			junk = IOObjectRelease(gCurrentService);
			assert(junk == noErr);
			gCurrentService = MACH_PORT_NULL;
			
			junk = IOObjectRelease(gCurrentServiceNotification);
			assert(junk == noErr);
			gCurrentServiceNotification = MACH_PORT_NULL;
		} else if (messageType == kIOMessageServiceIsRequestingClose) {
			if (gFWIntf != nil) {
				(*gFWIntf)->Close(gFWIntf);
				// how to free gFWIntf
				gFWIntf = nil;
			}
		} else if (messageType == kIOMessageServiceWasClosed) {
			gFWIntf = nil;
			gPlugIn = nil;
		} else {
			fprintf(stderr, "Unexpected message %0lx\n", messageType);
		}
	}
}

static void MyDeviceAddedCallback(void *refCon, io_iterator_t iterator)
{
	OSStatus junk;
	#pragma unused(refCon)
		
	Boolean 		done;
	long long		guid;

	// Look through all of the devices that have shown up.
	
	do {
		io_service_t 	thisService;
		CFNumberRef		guidRef;
		int				deviceIndex;
		
		thisService = IOIteratorNext(iterator);
		done = (thisService == MACH_PORT_NULL);
		if ( ! done ) {
			if (gCurrentService == MACH_PORT_NULL) {
				guidRef = IORegistryEntryCreateCFProperty(thisService, CFSTR("GUID"), nil, 0);
				if ( 	(guidRef != nil) 
					 && (CFGetTypeID(guidRef) == CFNumberGetTypeID()) 
					 && (CFNumberGetValue(guidRef, kCFNumberLongLongType, &guid)) ) {

					// If the device has a GUID property, search through our 
					// list of known devices for someone that we're interested.
					
					for (deviceIndex = 0; deviceIndex < (sizeof(gDevices) / sizeof(*gDevices)); deviceIndex++) {
						if (true && gDevices[deviceIndex].guid == guid) {
							if (FoundDevice(thisService, deviceIndex) == noErr) {
								gFirstTime = true;
								gCurrentService = thisService;
								thisService = MACH_PORT_NULL;
							}
						}
					}
				}
				CFQRelease(guidRef);
			}
		}
		
		if (thisService != MACH_PORT_NULL) {
			junk = IOObjectRelease(thisService);
			assert(junk == 0);
		}
	} while ( ! done );
}

static OSStatus InitFireWireStuff(void)
{
	OSStatus err;
	OSStatus junk;
	CFMutableDictionaryRef 	matchDict;

	matchDict = nil;
	
	err = IOMasterPort(MACH_PORT_NULL, &gMasterPort);
	if (err == noErr) {
		matchDict = IOServiceMatching("IOFireWireDevice");
		if (matchDict == NULL) {
			err = -1;
		}
	}
	if (err == noErr) {
		gNotificationPortRef = IONotificationPortCreate(gMasterPort);
		if (gNotificationPortRef == nil) {
			err = -1;
		}
	}
	if (err == noErr) {
		CFRunLoopAddSource(CFRunLoopGetCurrent(), 
						   IONotificationPortGetRunLoopSource(gNotificationPortRef), 
						   kCFRunLoopCommonModes);
	}
	if (err == noErr) {
		err = IOServiceAddMatchingNotification(gNotificationPortRef, 
											   kIOFirstPublishNotification,
											   matchDict, 
											   MyDeviceAddedCallback, nil, 
											   &gIterator);
		matchDict = nil;
	}
	if (err == noErr) {
		MyDeviceAddedCallback(nil, gIterator);
	}
	if (gIterator != MACH_PORT_NULL) {
		junk = noErr;
		// junk = IOObjectRelease(gIterator);
		assert(junk == noErr);
	}

	CFQRelease(matchDict);
	
	return err;
}

int main(void)
{
	kern_return_t 			err;
	
	err = Initialize();
	if (err == noErr) {
		err = InitFireWireStuff();
	}
	if (err == noErr) {
		RunApplicationEventLoop();
	}
	
	return 0;
}
